<?php

namespace Drupal\rest_log\EventSubscriber;

use Drupal\Component\Datetime\TimeInterface;
use Drupal\Component\Serialization\Json;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityStorageException;
use Drupal\Core\Routing\CurrentRouteMatch;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\rest\ResourceResponse;
use Drupal\rest_log\Entity\RestLog;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;
use Symfony\Component\HttpKernel\Event\ResponseEvent;
use Symfony\Component\HttpKernel\Event\TerminateEvent;
use Symfony\Component\HttpKernel\KernelEvents;
use Symfony\Component\Serializer\Encoder\XmlEncoder;

/**
 * Logs all rest responses/exceptions.
 */
class RestLogSubscriber implements EventSubscriberInterface {

  use StringTranslationTrait;

  /**
   * The exception if exists.
   *
   * @var \Throwable|null
   */
  protected $throwable;

  /**
   * The current route match.
   *
   * @var \Drupal\Core\Routing\CurrentRouteMatch
   */
  protected $currentRouteMatch;

  /**
   * @var \Drupal\Core\Config\ConfigFactoryInterface
   */
  private $configFactory;

  /**
   * @var \Drupal\Component\Datetime\TimeInterface
   */
  private TimeInterface $time;

  /**
   * @var \SplStack
   */
  private \SplStack $stack;

  /**
   * Creates rest_log event subscriber.
   */
  public function __construct(
    CurrentRouteMatch $current_route_match,
    ConfigFactoryInterface $config_factory,
    TimeInterface $time
  ) {
    $this->currentRouteMatch = $current_route_match;
    $this->configFactory = $config_factory;
    $this->time = $time;
    $this->stack = new \SplStack();
  }

  /**
   * Whether to log the requests with same-host referrer.
   *
   * @return bool
   *   TRUE if the requests with same-host referrer should be logged.
   */
  private function includeSameHost(): bool {
    return $this->configFactory->get('rest_log.settings')->get('include_same_host');
  }

  /**
   * Check if requested path must be logged.
   *
   * @return bool
   *   True if request is REST.
   */
  protected function isRestPage(): bool {
    $route = $this->currentRouteMatch->getRouteObject();
    return $route && $route->hasDefault('_rest_resource_config');
  }

  /**
   * Encoding password.
   *
   * @param string $password
   *   Password of type string.
   *
   * @return string
   *   return string.
   */
  public function encodePassword(string $password) {
    $substitutor = '***';
    if (strlen($password) <= 6) {
      $pass = substr($password, 0, 3) . $substitutor;
    }
    else {
      $pass = substr($password, 0, 3) . $substitutor . substr($password, -3);
    }
    return $pass;
  }

  /**
   * Log the REST response.
   *
   * @param \Symfony\Component\HttpKernel\Event\ResponseEvent $event
   *   The response event.
   */
  public function logResponse(ResponseEvent $event): void {
    if (!$this->isRestPage()) {
      return;
    }
    $request = $event->getRequest();
    $response = $event->getResponse();
    // Do not log requests with same-host referrer if disabled.
    if ($request->getHost() === parse_url($request->headers->get('referer'), PHP_URL_HOST) && !$this->includeSameHost()) {
      return;
    }
    $this->stack->push([
      'request' => $request,
      'response' => $response,
    ]);
  }

  /**
   * Log the REST response.
   *
   * @param \Symfony\Component\HttpKernel\Event\TerminateEvent $event
   *   The response event.
   */
  public function terminate(TerminateEvent $event): void {
    foreach ($this->stack as $item) {
      $this->doLogResponse($item['request'], $item['response']);
    }
  }

  /**
   * Log the REST response.
   *
   * @param \Symfony\Component\HttpFoundation\Request $request
   *   The request.
   * @param \Symfony\Component\HttpFoundation\Response $response
   *   The response.
   */
  private function doLogResponse($request, $response) {
    // Prepare the response log (in json format).
    $response_content_type = $response->headers->get('Content-type');
    $response_content = $response->getContent();
    switch ($response_content_type) {
      case 'application/octet-stream':
        // Do not log files.
        $responseBody['Content'] = 'File';
        break;
      case 'text/xml':
        $responseBody = (new XmlEncoder)->decode($response_content, 'array');
        break;
      default:
        $responseBody = Json::decode($response_content);
        break;
    }

    if ($this->throwable) {
      $responseBody['Exception'] = $this->throwable->getMessage();
    }

    // Prepare the request log.
    $request_content_type = $request->headers->get('Content-type');
    $request_content = $request->getContent();
    if ($request_content_type === 'application/octet-stream') {
      // Do not log files.
      $request_content = '';
    }

    $time = round(1000 * (microtime(TRUE) - $this->time->getRequestMicroTime()));

    $responseLog = array_filter([
      'request_header' => print_r($this->cleanUpHeaders($request->headers->all()), 1),
      'request_method' => $request->getMethod(),
      'request_payload' => $request_content,
      'response_status' => $response->getStatusCode(),
      'response_header' => print_r($this->cleanUpHeaders($response->headers->all()), 1),
      'response_body' => Json::encode($responseBody),
      'request_uri' => $request->getUri(),
      'request_cookie' => print_r($this->cleanUpCookies($request->cookies->all()), 1),
      'response_time' => $time,
    ]);

    try {
      RestLog::create($responseLog)->save();
    }
    catch (EntityStorageException $exception) {
      watchdog_exception('rest_log', $exception);
    }
  }

  /**
   * On rest exception log the event and return clear message.
   *
   * @param \Symfony\Component\HttpKernel\Event\ExceptionEvent|\Symfony\Component\HttpKernel\Event\ExceptionEvent $event
   *   The exception event.
   */
  public function onException($event): void {
    if (!$this->isRestPage()) {
      return;
    }
    $this->throwable = $event->getThrowable();
    $errorResponse = [
      'status' => 'error',
      'message' => $this->t('System error, please contact the administrator'),
    ];
    $response = new ResourceResponse($errorResponse);
    $event->setResponse($response);
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    $events = [];
    $events[KernelEvents::RESPONSE][] = ['logResponse', 1000];
    $events[KernelEvents::EXCEPTION][] = ['onException', -254];
    $events[KernelEvents::TERMINATE][] = ['terminate', 1000];
    return $events;
  }

  /**
   * Transform header arrays as arrays to key-value and sanitize.
   *
   * @param array $headers
   *
   * @return array
   */
  protected function cleanUpHeaders(array $headers): array {
    foreach ($headers as $headerKey => $header) {
      if (is_array($header) && count($header) === 1) {
        $headers[$headerKey] = $header[0];
      }
      // Remove authorization header since it's potential security issue.
      if (in_array(strtolower($headerKey), [
        'authorization',
        'php-auth-user',
        'php-auth-pw',
        'PHP_AUTH_PW',
        '_password',
        'cookie',
      ])) {
        $pass_code = $headers[$headerKey];
        if ($pass_code && $headerKey == 'authorization') {
          [$authscheme, $credentials] = preg_split('/\s+/', $pass_code);
          $headers[$headerKey] = $authscheme . ' ' . $this->encodePassword($credentials);
        }
        else {
          $headers[$headerKey] = $this->encodePassword($pass_code);
        }
      }
    }
    return $headers;
  }

  /**
   * Transform cookie arrays as arrays to key-value and sanitize.
   *
   * @param array $all
   *
   * @return array
   */
  protected function cleanUpCookies(array $all): array {
    foreach ($all as $key => $value) {
      if (
        str_starts_with($key, 'SESS')
        || str_starts_with($key, 'SSESS')
        || str_starts_with($key, 'openid_connect')
      ) {
        $all[$key] = $this->encodePassword($value);
      }
    }
    return $all;
  }

}
